/************************************************************************************
* Copyright (C) 1999 Cristiano Sadun crsadun@tin.it
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
************************************************************************************/

package com.deltax.cpacker;

import java.awt.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.zip.*;
import java.util.jar.*;
import javax.swing.*;
import org.w3c.tools.crypt.Md5;

import com.deltax.util.*;

/**
 * This class checks and packs togheter in a JAR file all the .class files
 * necessary to resolve, link and execute a given Java class.
 * <br>
 * A ClassNameFilter object is used to filter out unwanted classes.
 * By default, a <b>JavaNameFilter</b> is used, which excludes both
 * java.* and javax.* classes.
 * <br>
 * A ClassFileFinder object is used to locate the class files for
 * building the JAR. By default, the JDK 1.2 class loading scheme is
 * emulated to retrieve the necessary files, but an arbitrary ClassFileFinder
 * can be set if diverse class find schemes are necessary.
 * <br>
 * A FileNameProvider object can be optionally used to specify a file
 * name for the method <tt>createJar</tt>.
 * By default, a DefaultFileNameProvider, which prompts the user for
 * a file name is used.
 *
 * @author Cristiano Sadun
 * @version 1.4
 */
public class ClassPacker {

   private Hashtable set = new Hashtable();
   private ClassNameFilter filter;
   private Class cls;
   private Vector scanListeners = new Vector();
   private ClassFileFinder cff = new JDK12ClassFileFinder();
   private FileNameProvider fnn = new DefaultFileNameProvider();
   private RmiImplNameProvider rnp = new DefaultRmiImplNameProvider();
   private CPoolReader cpr = new CPoolReader();
   
   private Hashtable forNameUsers = new Hashtable();
   private boolean scanForNameUsage=false;
   private boolean jarAppend=false;
   private boolean incRMISkel=false;
   private boolean incRMIStub=false;
   private boolean createManifestFile=true;
   private String mainClass; // Manifest file main class
   
   private static Class javaRmiRemoteCls;

   public ClassPacker(String name) throws ClassNotFoundException {
      this(name, null);
   }

   public ClassPacker(String name, ClassNameFilter filter) throws ClassNotFoundException {
      this(Class.forName(name), filter);
   }

   public ClassPacker(Class cls, ClassNameFilter filter) {
      this.filter=filter;
      this.cls=cls;
   }

   public ClassPacker(Class cls) {
      this(cls, null);
   }

   public ClassFileFinder getClassFileFinder() { return cff; }
   public void setClassFileFinder(ClassFileFinder cff) { this.cff=cff; }
   public ClassNameFilter getClassNameFilter() { return filter; }
   public void setClassNameFilter(ClassNameFilter filter) { this.filter=filter; }
   public FileNameProvider getFileNameProvider() { return fnn; }
   public void setFileNameProvider(FileNameProvider fnn) { this.fnn=fnn; }
   public void setRmiImplNameProvider(RmiImplNameProvider rnp) { this.rnp=rnp; }
   public void setAppend(boolean v) { jarAppend = v; }
   public boolean getAppend() { return jarAppend; }
   public boolean getCreateManifestFile() { return createManifestFile; }
   public void setCreateManifestFile(boolean v) { createManifestFile=v; }
   public String getManifestMainClass() { return mainClass; }
   public void setManifestMainClass(String c) { mainClass=c; }
   

   public void addScanListener(ClassScanListener l) {
      if (! scanListeners.contains(l)) scanListeners.addElement(l);
   }
   public void removeScanListener(ClassScanListener l) {
      scanListeners.removeElement(l);
   }

   /**
    * After a successful scanning, returns the names of the
    * classes set
    * @return an enumeration of String objects
    */
   public Enumeration getNames() { return set.keys(); }

   /**
    * After a successful scanning, returns the classes set
    * @return an enumeration of Class objects
    */
   public Enumeration getClasses() { return set.elements(); }
   public int countClasses() { return set.size(); }

   /**
    * This inteface defines a class name filter.
    */
   public interface ClassNameFilter {
      /**
       * @return true if the className is accepted, false otherwise
       */
      public boolean accept(String className);
   }

   /**
    * This inteface must be implemented by any object interested
    * in listening to scanning events
    */
   public interface ClassScanListener {
      /**
       * Called when a class is added.
       * This is called in the same thread as the ClassPacker object.
       */
      public void classAdded(String className);
      
      /**
       * Called when a not strictly fatal exeception occurs during scanning.
       * Implementor can decide to abort or continue.
       * This is called in the same thread as the ClassPacker object.
       */
      public void nonFatalException(String msg, Exception e);

      /**
       * Called when the scan is complete.
       * This is called in the same thread as the ClassPacker object.
       */
      public void scanEnd();
   }

   /**
    * This inteface must be implemented by any class providing
    * a file name for the JAR file constructed by createJAR
    */
   public static interface FileNameProvider {
      
      public static class CanceledException extends Exception {
      }
   
      /**
       * Called by each invocation of createJAR()
       */
      public String getFileName() throws CanceledException;
   }
   
   /**
    * This interface must be implemented by any class providing
    * RMI implementation name conventions for searching implementations
    * of Remote interfaces.
    */
   public static interface RmiImplNameProvider {
   	/**
   	 * Given an interface name, an array containing all the possible names for
   	 * implementation skeleton (built following some convention defined by the 
   	 * provider) is returned.
   	 */
   	public String [] getImplSkelNames(String interfaceName);
   	
   	/**
   	 * Given an interface name, an array containing all the possible names for
   	 * implementation stub (built following some convention defined by the 
   	 * provider) is returned.
   	 */
   	public String [] getImplStubNames(String interfaceName);
   }

   /**
    * This class filters out "java.*" and (optionally) "javax.*"
    * class names.
    */
   public static class JavaNameFilter implements ClassNameFilter {

      private boolean excludeExtensions;

      /**
       * Builds a filter which filters out "java.*" and "javax.*"
       * class names.
       */
      public JavaNameFilter() {
         this(true);
      }

      /**
       * Builds a filter which filters out "java.*" and allows to
       * set filtering for "javax.*" class names.
       * @param excludeExtensions true if "javax.*" names have to be
       *        filtered out
       */
      public JavaNameFilter(boolean excludeExtensions) {
         this.excludeExtensions = excludeExtensions;
      }

      public boolean accept(String className) {
         if (className.startsWith("java.")) return false;
         if (excludeExtensions)  if (className.startsWith("javax.")) return false;
         return true;
      }
   }

   /**
    * This class prompts a GUI to ask the user for a file name.
    * No validation is made upon the file name.
    */
   public static class DefaultFileNameProvider implements FileNameProvider {

      private String fname;

      public String getFileName() throws FileNameProvider.CanceledException {
         if (fname==null) fname=JOptionPane.showInputDialog("JAR File name?");
         if (fname==null) throw new FileNameProvider.CanceledException();
         return fname;
      }
      public void setFileName(String fname) { this.fname=fname; }

	}

   /**
    * Defines the default conventions for searching implementation classes of Remote
    * interfaces.
    * <p>
    * A default rmi implementation-name provider which defines that a 
    * class implementing a Remote interface P1.P2.I will have as possible names:
    * <ul>
    * <li>&lt;I&gt;Impl in the same package P1.P2 as I
    * <li>&lt;I&gt;Server in the same package P1.P2 as I
    * <li>&lt;I&gt;in the package P1.P2.server
    * <li>&lt;I&gt;in the package P1.P2.impl
    * <li>&lt;I&gt;Server in the package P1.P2.server
    * <li>&lt;I&gt;Impl in the package P1.P2.server
    * <li>&lt;I&gt;Server in the package P1.P2.impl
    * <li>&lt;I&gt;Impl in the package P1.P2.impl
    * <li>&lt;I&gt;Server in the package P1.server (if exists)
    * <li>&lt;I&gt;Impl in the package P1.server (if exists)
    * <li>&lt;I&gt;Server in the package P1.impl (if exists)
    * <li>&lt;I&gt;Impl in the package P1.impl (if exists)
    * </ul>
    */
	public static class DefaultRmiImplNameProvider implements RmiImplNameProvider {
		
		private String [] getImplNames(String interfaceName, String postFix) {
			int i=interfaceName.lastIndexOf(".");
			String pkgName;
			String simpleName;
			if (i!=-1) { 
				pkgName=interfaceName.substring(0, i+1);
				simpleName=interfaceName.substring(i+1);
			} else { 
				pkgName=""; 
				simpleName=interfaceName; 
			}
			
			// Build names array
			Vector names = new Vector();
			names.addElement(pkgName+simpleName+"Impl"+postFix);
			names.addElement(pkgName+simpleName+"Server"+postFix);
			names.addElement(pkgName+"server."+simpleName+postFix);
			names.addElement(pkgName+"impl."+simpleName+postFix);
			names.addElement(pkgName+"server."+simpleName+"Server"+postFix);
			names.addElement(pkgName+"impl."+simpleName+"Server"+postFix);
			names.addElement(pkgName+"server."+simpleName+"Impl"+postFix);
			names.addElement(pkgName+"impl."+simpleName+"Impl"+postFix);
			
			i=pkgName.substring(0, pkgName.length()-1).lastIndexOf(".");
			if (i!=-1) {
				String p1 = pkgName.substring(0,i+1);
				names.addElement(p1+"server."+simpleName+"Server"+postFix);
				names.addElement(p1+"impl."+simpleName+"Server"+postFix);
				names.addElement(p1+"server."+simpleName+"Impl"+postFix);
				names.addElement(p1+"impl."+simpleName+"Impl"+postFix);
			}
			
			String [] nms = new String[names.size()];
			names.copyInto(nms);
			return nms;
		}
		
   	public String [] getImplSkelNames(String interfaceName) {
   		return getImplNames(interfaceName, "_Skel");
	   }
   
   	public String [] getImplStubNames(String interfaceName) {
   		return getImplNames(interfaceName, "_Stub");
   	}
   }
   
   /** 
    * A class which uses property files to determine implementations for RMI Remote 
    * interfaces.
    *
    * The property file must contain couples:
    * <p>
    * <tt>interface name : implementation class name</tt>
    * <p>
    * as in
    * <pre>
    * # RMI Implementation definitions for MyApplication
    * #
    * com.deltax.util.session.SessionManager : com.deltax.util.session.impl.SessionManagerImpl
    * com.deltax.myapp.HelloObject : com.deltax.myapp.HelloObjectImpl
    * </pre>
    *
    * If no property file is specified (passing null as filename on construction), 
    * the same conventions as DefaultImplNameProvider are used. 
    *
    * However, if a file is specified and it does not contain an implementation definition
    * for a given class, the class is reported as missing.
    * 
    */
   public static class PropertiesImplNameProvider implements RmiImplNameProvider {
   	
   	private DefaultRmiImplNameProvider helper = new DefaultRmiImplNameProvider();
   	private Properties prop;
   	
   	/**
   	 * Attempts to load the given property file which defines the associations
   	 * between Remote interfaces and their implementations
   	 */
   	public PropertiesImplNameProvider (String propertiesFile) throws IOException {
   		if (propertiesFile!=null) {
   			 prop = new Properties();
   			 prop.load(new FileInputStream(propertiesFile));
   		}
   	}
   	
   	public String [] getImplSkelNames(String interfaceName) {
   		if (prop==null) return helper.getImplSkelNames(interfaceName);
   		String impCls = prop.getProperty(interfaceName);
   		if (impCls==null) return new String[0];
   		String [] res = new String[1];
   		res[0] = impCls+"_Skel";
   		return res;
	   }
   
   	public String [] getImplStubNames(String interfaceName) {
   		if (prop==null) return helper.getImplSkelNames(interfaceName);
   		String impCls = prop.getProperty(interfaceName);
   		if (impCls==null) return new String[0];
   		String [] res = new String[1];
   		res[0] = impCls+"_Stub";
   		return res;
   	}
   }

   /**
    * This class simply pops up a GUI Frame (Swing) to show the
    * scanning status.
    */
   public static class Monitor implements ClassScanListener {

      JFrame dlg;
      JLabel l;
      ClassPacker cp;

      public Monitor(ClassPacker cp) {
         this(cp, null);
      }

      public Monitor(ClassPacker cp, Component parent/*, boolean modal*/) {
         this.cp = cp;
         dlg = new JFrame("ClassPacker 1.4 (C) Copyright 1999 Cristiano Sadun - crsadun@tin.it");

         Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
         dlg.setLocation((d.width-dlg.getSize().width)/2, (d.height-dlg.getSize().height)/2);
         l=new JLabel("  Scanning..");
         dlg.getContentPane().setLayout(new BoxLayout(dlg.getContentPane(), BoxLayout.Y_AXIS));
         dlg.getContentPane().add(new JLabel("  Scanning classes.."));
         dlg.getContentPane().add(Box.createVerticalStrut(5));
         dlg.getContentPane().add(l);
         dlg.getContentPane().add(Box.createVerticalStrut(10));
         JLabel lb;
         Font f=new Font("Times", Font.PLAIN, 10);
         dlg.getContentPane().add(lb=new JLabel(" Part of this software is provided under"));
         lb.setFont(f);
         lb.setForeground(Color.black);
         dlg.getContentPane().add(lb=new JLabel(" the GNU Public License Library"));
         lb.setFont(f);
         lb.setForeground(Color.black);
         dlg.getContentPane().add(lb=new JLabel(" in both source and executable format"));
         lb.setFont(f);
         lb.setForeground(Color.black);
         dlg.getContentPane().add(lb=new JLabel(" at http://space.tin.it/computer/csadun/software/cpacker/index.html"));
         lb.setFont(f);
         lb.setForeground(Color.black);
         dlg.getContentPane().add(Box.createVerticalStrut(5));
         dlg.pack();
         //dlg.setSize(350,80);
         dlg.setResizable(false);
         dlg.setVisible(true);
      }

      public void classAdded(String className) {
         l.setText("  Added class "+className);
         l.updateUI();
      }
      
      public void nonFatalException(String msg, Exception e) {
      	System.out.println("Warning: "+msg);
      }

      public void scanEnd() {
         cp.removeScanListener(this);
         dlg.setVisible(false);
         dlg.dispose();
         l=null;
         dlg=null;
      }
   }

   /**
    * Starts scanning the class. Each addition is notified to the
    * listeners (if any), as well as the end of the process.
    */
   public void startScan() {
      makeSet(cls);
      notifyEnd();
   }


   /**
    * Creates a JAR file for the collected classes with the given name
    * @param jarfile the JAR file name. If null, the current FileNameProvider will be used
    * @exception ClassNotFoundException if the class is not found. See class
                 description for a discussion of class finding
    * @exception IOException if an I/O exception occurs
    */
   public void createJar(final String jarfile) throws ClassNotFoundException, IOException {
      if (jarfile != null)
         setFileNameProvider(new FileNameProvider() {
            public String getFileName() { return jarfile; }
         }
         );
      createJar();
   }
   
   public void setScanForNameUsage(boolean v) { scanForNameUsage = v; }
   public boolean getScanForNameUsage() { return scanForNameUsage; }
   public void setIncRMIStub(boolean v) { incRMIStub = v; }
   public void setIncRMISkel(boolean v) { incRMISkel = v; }

   /**
    * Creates a JAR file for the collected classes.
    * The current FileNameProvider is invoked to obtain a file name
    * for the JAR file.
    */
   public void createJar() throws ClassNotFoundException, IOException {

      try {
         if (new File(fnn.getFileName()).exists() && (! jarAppend)) {
            if (JOptionPane.showConfirmDialog(
               null,
               "File \""+fnn.getFileName()+"\" already exists. Do you want to overwrite it?",
               "File exists",
               JOptionPane.YES_NO_OPTION)==JOptionPane.YES_OPTION)
                  new File(fnn.getFileName()).delete();
            else return;
         }
   
         
         Enumeration en = getClasses();
         PrintWriter manifestWriter=null;
         StringWriter manifestStringWriter=null;
         
         // Create a manifest in memory, if required
         if (createManifestFile) {
         	manifestWriter=new PrintWriter((manifestStringWriter=new StringWriter()));
         	manifestWriter.println("Manifest-Version: 1.0");
         	manifestWriter.println("Created-By: ClassPacker 1.4 (C) 1999,2000 dr. Cristiano Sadun, crsadun@tin.it");
         	if (mainClass!=null) manifestWriter.println("Main-Class: "+mainClass);
         	manifestWriter.println();
         }
         
         JarOutputStream jos = new JarOutputStream(new FileOutputStream(fnn.getFileName(), jarAppend));
         Vector dirs = new Vector();   // Directories created in zipfile
   
         ProgressMonitor pm = new ProgressMonitor(null, "Creating JAR file "+fnn.getFileName(), "Initializing", 0, 2*countClasses());
   
         int count = 1;
         while (en.hasMoreElements()) {
            Class c = (Class)en.nextElement();
   
            // Locate and open class
            pm.setNote("Locating "+c.getName());
            pm.setProgress(++count);
            if (pm.isCanceled()) { abort(fnn.getFileName(), jos); break; }
   
            InputStream is = getClassFileFinder().openClass(c.getName());
   
            // Transform class name in Zip path
            String name = c.getName().replace('.','/')+".class";
            // Check for directory entries
            if (name.lastIndexOf("/") != -1 ) {
               String dir = name.substring(0, name.lastIndexOf("/")+1);
               if (! dirs.contains(dir)) {
                  ZipEntry ze = new ZipEntry(dir);
                  jos.putNextEntry(ze);
                  jos.closeEntry();
                  dirs.addElement(dir);
               }
            }
   
            ZipEntry entry = new ZipEntry(name);
            jos.putNextEntry(entry);
            
            // Add the name to the manifest, if required
            if (createManifestFile) {
            	manifestWriter.print("Name: ");
            	manifestWriter.println(name);
            	manifestWriter.println("Digest-Algorithms: MD5");
            	Md5 md5 = new Md5(name);
            	md5.processString();
            	manifestWriter.print("MD5-Digest: ");
            	manifestWriter.println(md5.getStringDigest());
            	manifestWriter.println();
            }
   
            // Load class data and put it in JAR file
            pm.setNote("Adding "+c.getName());
            pm.setProgress(++count);
            if (pm.isCanceled()) { abort(fnn.getFileName(), jos); break; }
   
            int ch;
            while ((ch=is.read()) != -1) jos.write(ch);
            is.close();
            jos.closeEntry();
         }
         //System.out.println(count/2+" classes added");
         
         // Create and add the Manifest File, if required
         if (createManifestFile) {
         	pm.setNote("Adding Manifest file");
         	manifestWriter.close();
         	ZipEntry entry = new ZipEntry("META-INF/");
         	jos.putNextEntry(entry);
         	jos.closeEntry();
         	entry = new ZipEntry("META-INF/MANIFEST.MF");
         	jos.putNextEntry(entry);
            Reader r = new StringReader(manifestStringWriter.toString());
            int ch;
            while ((ch=r.read()) != -1) jos.write(ch);
            r.close();
            jos.closeEntry();
         }
         
         jos.close();
         pm.close();
      } catch (FileNameProvider.CanceledException e) { 
         // This can be fired only at the start with the default FNP
         // and no action is required
      }
   }

   /**
    * Creates a report to the given printstream for classes which use Class.forName
    * (and are therefore likely to need manual additions to the JAR file)
    */
   public void createForNameUsageReport(String clsName, PrintStream p) {
         Class [] cls = getForNameUserClasses();
         p.println("# ClassPacker v1.2 (C) 1999 Cristiano Sadun crsadun@tin.it");
         p.println("#");
         p.println("# Classes using Class.forName(String clsName) method");
         p.println("# found while packing "+clsName);
         p.println("# on "+new Date());
         p.println("#");
         p.println("# (This file is generated automatically)");
         for(int i=0;i<cls.length;i++) {
            p.println(cls[i].getName());
         }
   }
   
   private void abort(String jarfile, JarOutputStream jos) {
      try {
         jos.close();
         new File(jarfile).delete();
      } catch(IOException e) {
         e.printStackTrace();
      }
   }

   private void notifyAdd(Class c) {
      Enumeration e = scanListeners.elements();
      while (e.hasMoreElements()) {
         ClassScanListener l = (ClassScanListener)e.nextElement();
         l.classAdded(c.getName());
      }
   }

   private void notifyEnd() {
      Enumeration e = scanListeners.elements();
      while (e.hasMoreElements()) {
         ClassScanListener l = (ClassScanListener)e.nextElement();
         l.scanEnd();
      }
   }
   
   private void notifyException(String msg, Exception exc)  {
   	Enumeration e = scanListeners.elements();
      while (e.hasMoreElements()) {
         ClassScanListener l = (ClassScanListener)e.nextElement();
         l.nonFatalException(msg, exc);
      }
   }

   private void makeSet(Class cls []) {
      for(int i=0;i<cls.length;i++)
         makeSet(cls[i]);
   }

   private void makeSet(Field [] fields) {
      for(int i=0;i<fields.length;i++)
         makeSet(fields[i].getType());
   }

   private void makeSet(Method [] methods) {
      for(int i=0;i<methods.length;i++) {
         makeSet(methods[i].getReturnType());
         makeSet(methods[i].getParameterTypes());
         makeSet(methods[i].getExceptionTypes());
      }
   }
   
   /**
    * This is the main makeSet function, which inspects the class updating the
    * set.
    * We have a problem here: reflection is not powerful enough to
    * indicate all the *used* classes. So we need to read the constant pool
    * and extract the named classes therein.
    * Since Class does not provide services to revert to bytecode, we actually
    * need to locate the class to get the byte stream, and inspect there.
    * This slows down things - each class bytecode gets read twice - but that's
    * it for the moment, unless a Class loader which allows direct call of
    * defineClass is built etc.
    * We still use reflection for what we can 'cause it's by far easier than
    * looking directly in the constant pool.
    */
   private void makeSet(Class c) {
   	
      // Use filter if available
      if (filter!=null)
            if (! filter.accept(c.getName())) return;

      // Primitive types are not interesting
      if (c.isPrimitive()) return;

      // Exclude if the class is already in set (doing this *before* recursing
      // avoids loopholes besides duplicates)
      if (set.containsKey(c.getName())) return;

      // If it's an array, add the declared type
      if (c.isArray()) {
         makeSet(c.getComponentType());
         return;
      }
      
      // Add itself
      set.put(c.getName(), c);
      notifyAdd(c);
      if (scanForNameUsage==true)
         if (isUsingClassForName(c)) forNameUsers.put(c.getName(), c);

      // Used classes
      makeSet(c.getClasses());

      // Constructor exceptions
      Constructor cn[]=c.getConstructors();
      for (int i=0;i<cn.length;i++)
         makeSet(cn[i].getExceptionTypes());

      // Declared classes/interfaces
      makeSet(c.getDeclaredClasses());

      // Member classes
      Field [] fields = c.getDeclaredFields();
      makeSet(fields);

      // Methods return classes and parameter classes
      Method [] methods = c.getDeclaredMethods();
      makeSet(methods);

      // Implemented interfaces
      Class [] interfaces = c.getInterfaces();
      makeSet(interfaces);
      
      // Extended class
      if (c.getSuperclass() != null)
         makeSet(c.getSuperclass());
      
      // If the class extends the Remote interface (and is not *the* remote interface interface), check
      // if the stubs or skels have to be added
      if (((incRMISkel || incRMIStub) && javaRmiRemoteCls.isAssignableFrom(c))) {
      	//System.out.println(c.getName()+" is assignable from java.rmi.Remote");
      	if( !("java.rmi.Remote".equals(c.getName())) && c.isInterface()) {
	         //System.out.println(c.getName()+" implements java.rmi.Remote");
				if (incRMISkel) addRmiImplementations(c, true);
				if (incRMIStub) addRmiImplementations(c, false);
      	}
      }
      
      // Used classes - most of them will be already included..
      makeSet(getUsedClasses(c));
   }
   
   private void addRmiImplementations(Class c, boolean skeleton) {
		// Search the standard names
		String [] namesToSearch = (skeleton ?
			rnp.getImplSkelNames(c.getName()) :
			rnp.getImplStubNames(c.getName()));
			
		/*
		System.out.println("Searching for:");
		for(int j=0;j<namesToSearch.length;j++) 
			System.out.println(namesToSearch[j]);
		*/
			
		Class implCls = null;
		for(int j=0;j<namesToSearch.length;j++) 
			try {
				implCls = Class.forName(namesToSearch[j]);
				// The class must implement the remote interface, not just have the right name
				if (c.isAssignableFrom(implCls)) break;
			} catch (ClassNotFoundException e) { }

		if (implCls==null) { // No implementation found
			StringBuffer sb = new StringBuffer("No class implementing the RMI interface");
			sb.append(c.getName());
			sb.append("(Searched: ");
			sb.append("\n");
			for(int j=0;j<namesToSearch.length;j++) {
				sb.append("    ");	
				sb.append(namesToSearch[j]);
				if (j<namesToSearch.length-1) sb.append("\n");
			}
			sb.append(")");
			notifyException(sb.toString(), new ClassNotFoundException("Skeleton for "+c.getName()));
		} else {
			makeSet(implCls);
		}
      		
	}
   
   

   private Class[] getUsedClasses(Class c) {
      // Warning: this may fail if an odd classloader is used
      // Check the FileNameProvider in case of failure
      try {
         CPoolReader.classfile cf = cpr.readClass(c.getName());
         
         String [] names = cpr.getUsedClasses(c.getName());
         Class [] cls = new Class[names.length];
         for(int i=0;i<names.length;i++)
            cls[i]=Class.forName(names[i]);
            
         return cls;
      } catch(Exception e) {
      	e.printStackTrace();
         System.err.println(e.getClass().getName()+" : "+e.getMessage());
         System.err.println(
            "The class "+c.getName()+" bytecode cannot be located.\n"+
            "Remember that classes must be locable by using the Java "+cff.getSupportedLoadingScheme()+
            "class loading scheme");
         return new Class[0];
         //throw new RuntimeException("Packer failed");
      }
   }
   
   private Class [] getForNameUserClasses() {
      
      Class [] cls = new Class[forNameUsers.size()];
      Enumeration e = forNameUsers.elements();
      int i=0;
      while (e.hasMoreElements()) 
         cls[i++] = (Class)e.nextElement();
      return cls;
   }
   
   private boolean isUsingClassForName(Class c) {
      // Warning: this may fail if an odd classloader is used
      // Check the FileNameProvider in case of failure
      try {
         return cpr.forNameCalled(c.getName());
      } catch(Exception e) {
         System.err.println(e.getClass().getName()+" : "+e.getMessage());
         System.err.println(
            "The class "+c.getName()+" bytecode cannot be located.\n"+
            "Remember that classes must be locable by using the Java "+cff.getSupportedLoadingScheme()+
            "class loading scheme");
         return false;
         //throw new RuntimeException("Packer failed");
      }        
   }   

   private static String getBaseName(String clsName) {
      if (clsName.endsWith(".")) throw new RuntimeException("Invalid class name: "+clsName);
      if (clsName.lastIndexOf(".")==-1) return clsName;
      return clsName.substring(clsName.lastIndexOf(".")+1);
   }

   private static void usage() {
      System.out.println();
      System.out.println("ClassPacker r1.4");
      System.out.println("----------------------------------------------------------------------");
      System.out.println("(C) Copyright 1999, 2000 Cristiano Sadun - crsadun@tin.it");
      System.out.println();
      System.out.println("Part of this software is provided under the GNU Public License Library");
      System.out.println("in both source and executable format. More information can be found at");
      System.out.println();
      System.out.println("    http://space.tin.it/computer/csadun/software/cpacker/index.html");
      System.out.println();
      System.out.println("Usage:");
      System.out.println();
      System.out.println(" ClassPacker <fully qualified class name> [options]");
      System.out.println();
      System.out.println("   -o <JAR file name> uses name for output");
      System.out.println("   -ask  prompts the user for a JAR file name");
      System.out.println("    (If not specified the simple class name +\".jar\" is used for the JAR file)");
      System.out.println("   -fnr  generates report on Class.forName() usage");
      System.out.println("   -so   the previous report is generated to stdout");
      System.out.println("    (default is forNameList.txt file)");
      System.out.println("   -ijx  includes javax.* classes");
      System.out.println("   -ij   includes necessary java.* and javax.* classes");
      System.out.println("   -append adds the entries to an existing JAR file");
      System.out.println("	  -zip  use zip extension instead of JAR");
      System.out.println("   -nmf  does not create a manifest file");
      System.out.println("   -mc [class] specify mainfest main (default is class under packing)");
      System.out.println("   -skel <property file|default> (*) if a class implements java.rmi.Remote,");
      System.out.println("                                     an attempt is done to include the RMI ");
      System.out.println("                                     Skeleton of an implementation class");
      System.out.println("   -stub <property file|default> (*) if a class implements java.rmi.Remote,");
      System.out.println("                                     an attempt is done to include the RMI ");
      System.out.println("                                     Stub of an implementation class");
      System.out.println("   -rmi  <property file|default> (*) as specifying -stub -skel");
      System.out.println();
      System.out.println("   (*) the implmentations are searched for following the naming conventions");
      System.out.println("       described in the manual.");
      
      
      System.out.println();
      System.exit(0);
   }


   public static void main(String args[]) throws Exception {
      
         String jarfile=null;
         String clsName=null;
         String rmiImplPropertyFile=null;
         
         boolean askOpt=false;
         boolean fnrOpt=false;
         boolean soOpt=false;
         boolean excludeJavaExt=true;
         boolean excludeJava=true;
         boolean append=false;
         String extension="jar";
         boolean incRmiStub=false;
         boolean incRmiSkel=false;
         boolean createMF=true;
         String mainClass=null;
         if (args.length == 0) usage();
         for (int i=0;i<args.length;i++) 
            if (args[i].startsWith("-")) {
                 if ("-ask".equals(args[i])) askOpt=true; else
                 if ("-fnr".equals(args[i])) fnrOpt=true; else
                 if ("-so".equals(args[i])) soOpt=true; else
                 if ("-ijx".equals(args[i])) excludeJavaExt=false; else
                 if ("-ij".equals(args[i])) excludeJava=false; else
                 if ("-append".equals(args[i])) append=true; else
                 if ("-nmf".equals(args[i])) createMF=false; else
                 if ("-zip".equals(args[i])) extension="zip"; else
                 if ("-mc".equals(args[i])) {
                 	if (i < args.length-1) 
	                 	if ((args[i+1].startsWith("-"))) 
	                 		mainClass="packed";
	                 	else {
	                 		try {
	                 			mainClass=args[++i];
	                 			Class.forName(mainClass);
	                 		} catch (ClassNotFoundException e) {
	                 			throw new RuntimeException("The specified main class \""+mainClass+"\" cannot be found");
	                 		}
	                 	}
	               else mainClass="packed";
                 } else
                 if ("-skel".equals(args[i])) { 
                 		incRmiSkel=true; 
                 		if (i < args.length-1 && (!args[i+1].startsWith("-"))) {
                 			if (rmiImplPropertyFile!=null && ! rmiImplPropertyFile.equals(args[i+1]))
                 					throw new RuntimeException("RMI property files specified more than once");
                 				if ("default".equals(args[i+1])) { i++; continue; }
                 				rmiImplPropertyFile=args[++i];
                 			} else throw new RuntimeException("File name or \"default\" required for -skel");
                 } else
                 if ("-stub".equals(args[i])) { 
                 		incRmiStub=true; 
                 		if (i < args.length-1 && (!args[i+1].startsWith("-"))) {
                 			if (rmiImplPropertyFile!=null && ! rmiImplPropertyFile.equals(args[i+1]))
                 					throw new RuntimeException("RMI property files specified more than once");
	                 			if ("default".equals(args[i+1])) { i++; continue; }
	                 			rmiImplPropertyFile=args[++i];
                 		} else throw new RuntimeException("File name or \"default\" required for -stub");
                 } else
                 if ("-rmi".equals(args[i])) { 
                 		incRmiSkel=true; 
                 		incRmiStub=true; 
                 		if (i < args.length-1 && (!args[i+1].startsWith("-"))) { 
                 			if (rmiImplPropertyFile!=null && ! rmiImplPropertyFile.equals(args[i+1]))
                 					throw new RuntimeException("RMI property files specified more than once");
                 				if ("default".equals(args[i+1])) { i++; continue; }
                 				rmiImplPropertyFile=args[++i];
                 		} else throw new RuntimeException("File name or \"default\" required for -rmi");
                 } else
                 if ("-o".equals(args[i])) {
                 		if (i < args.length-1 && (!args[i+1].startsWith("-"))) jarfile=args[++i];
                 		else throw new RuntimeException("File name required for -o");
               	} 
                 else usage();
             } else {
             	clsName=args[i]; 
             }
             
         if (clsName ==null) usage();
         
         if ("packed".equals(mainClass)) mainClass=clsName;
            
         if (! askOpt && jarfile==null) {
             jarfile=getBaseName(clsName)+"."+extension;
         } 
         

         UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
   
         ClassPacker cp = new ClassPacker(clsName, excludeJava ? new JavaNameFilter(excludeJavaExt) : null);
	
			// Set forname usage flag         	
         cp.setScanForNameUsage(fnrOpt);
         
         // Set append flag
         cp.setAppend(append);	
         
         // Set create-manifest flag
         cp.setCreateManifestFile(createMF);
         
         // Set main class for mainfest file
         cp.setManifestMainClass(mainClass);
         
         // Set include RMI stubs/skeleton flag
         cp.setIncRMIStub(incRmiStub);
         cp.setIncRMISkel(incRmiSkel);
         if (rmiImplPropertyFile!=null) 
         	cp.setRmiImplNameProvider(new PropertiesImplNameProvider(rmiImplPropertyFile));

         // A class name filter which excludes classes already in JAR is used
         // if append is required
         if (append) 
         	cp.setClassNameFilter(new JarFileNameFilter(cp.getClassNameFilter(), jarfile));

			// Add the monitor         	
         ClassPacker.Monitor monitor = new ClassPacker.Monitor(cp);
         cp.addScanListener(monitor);
         
         cp.startScan();
         
         if (append)
         	if (! cp.getClasses().hasMoreElements()) { 
         		System.out.println("No classes to put in JAR."); 
         		return; 
         	}
         
         cp.createJar(jarfile);
         
         if (fnrOpt) {
            PrintStream p = soOpt ? System.out : new PrintStream(new FileOutputStream("forNameList.txt"));
            cp.createForNameUsageReport(clsName, p);
            if (! soOpt) System.out.println("Class.forName() usage report written to forNameList.txt");
         }
      
      	try {
      		Thread.sleep(1000);
      	} catch(InterruptedException e) {
      			e.printStackTrace();
      	}
      	
         System.exit(0); // Stop unwanted GUI threads
   }
   
   static {
   	try {
   		javaRmiRemoteCls=Class.forName("java.rmi.Remote");
   	} catch (ClassNotFoundException e) {
   		throw new NoClassDefFoundError("java.rmi.Remote");
   	}
   }
}